# sage_setup: distribution = sagemath-coxeter3
# sage.doctest: optional - coxeter3
"""
Coxeter Groups implemented with Coxeter3
"""
# ****************************************************************************
#       Copyright (C) 2009-2013 Mike Hansen <mhansen@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.libs.coxeter3.coxeter import get_CoxGroup, CoxGroupElement
from sage.misc.cachefunc import cached_method

from sage.structure.unique_representation import UniqueRepresentation
from sage.structure.element_wrapper import ElementWrapper
from sage.structure.richcmp import richcmp
from sage.categories.coxeter_groups import CoxeterGroups
from sage.structure.parent import Parent

from sage.combinat.root_system.coxeter_matrix import CoxeterMatrix

from sage.rings.integer_ring import ZZ
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing


class CoxeterGroup(UniqueRepresentation, Parent):
    @staticmethod
    def __classcall__(cls, cartan_type, *args, **options):
        """
        TESTS::

            sage: from sage.libs.coxeter3.coxeter_group import CoxeterGroup
            sage: CoxeterGroup(['B',2])
            Coxeter group of type ['B', 2] implemented by Coxeter3
            sage: CoxeterGroup(CartanType(['B', 3]).relabel({1: 3, 2: 2, 3: 1}))
            Coxeter group of type ['B', 3] relabelled by {1: 3, 2: 2, 3: 1} implemented by Coxeter3
        """
        from sage.combinat.root_system.cartan_type import CartanType
        ct = CartanType(cartan_type)
        return super().__classcall__(cls, ct, *args, **options)

    def __init__(self, cartan_type):
        """
        TESTS::

            sage: from sage.libs.coxeter3.coxeter_group import CoxeterGroup
            sage: CoxeterGroup(['A',2])
            Coxeter group of type ['A', 2] implemented by Coxeter3

        As degrees and codegrees are not implemented, they are skipped in the
        testsuite::

            sage: to_skip = ['_test_degrees', '_test_codegrees']
            sage: TestSuite(CoxeterGroup(['A',2])).run(skip=to_skip)
        """
        category = CoxeterGroups()
        if cartan_type.is_finite():
            category = category.Finite()
        Parent.__init__(self, category=category)
        self._coxgroup = get_CoxGroup(cartan_type)
        self._cartan_type = cartan_type

    def _repr_(self):
        """
        EXAMPLES::

            sage: W = CoxeterGroup(['A', 3], implementation='coxeter3'); W      # indirect doctest
            Coxeter group of type ['A', 3] implemented by Coxeter3
            sage: W = CoxeterGroup(['A', 3, 1], implementation='coxeter3'); W
            Coxeter group of type ['A', 3, 1] implemented by Coxeter3
        """
        return "Coxeter group of type %s implemented by Coxeter3" % (self.cartan_type())

    def __iter__(self):
        """
        EXAMPLES::

            sage: W = CoxeterGroup(['A', 2], implementation='coxeter3')
            sage: list(W)
            [[], [1], [2], [1, 2], [2, 1], [1, 2, 1]]
        """
        for x in self._coxgroup:
            yield CoxeterGroup.Element(self, x)

    def cartan_type(self):
        """
        Return the Cartan type for this Coxeter group.

        EXAMPLES::

            sage: W = CoxeterGroup(['A', 3], implementation='coxeter3')
            sage: W.cartan_type()
            ['A', 3]
        """
        return self._cartan_type

    def index_set(self):
        """
        Return the index set for the generators of this Coxeter group.

        EXAMPLES::

            sage: W = CoxeterGroup(['A', 3], implementation='coxeter3')
            sage: W.index_set()
            (1, 2, 3)
            sage: C = CoxeterGroup(['A', 3,1], implementation='coxeter3')
            sage: C.index_set()
            (0, 1, 2, 3)
        """
        return self.cartan_type().index_set()

    def bruhat_interval(self, u, v):
        """
        Return the Bruhat interval between ``u`` and ``v``.

        EXAMPLES::

            sage: W = CoxeterGroup(['A', 3], implementation='coxeter3')
            sage: W.bruhat_interval([1],[3,1,2,3])
            [[1], [1, 2], [1, 3], [1, 2, 3], [1, 3, 2], [1, 2, 3, 2]]
        """
        u, v = self(u), self(v)
        return self._coxgroup.bruhat_interval(u.value, v.value)

    def cardinality(self):
        """
        Return the cardinality of this Coxeter group.

        EXAMPLES::

            sage: W = CoxeterGroup(['A', 3], implementation='coxeter3')
            sage: W.cardinality()
            24
        """
        return self._coxgroup.order()

    def one(self):
        """
        Return the identity element of this Coxeter group.

        EXAMPLES::

            sage: W = CoxeterGroup(['A', 3], implementation='coxeter3')
            sage: W.one()
            []
        """
        return self.element_class(self, [])

    def simple_reflections(self):
        """
        Return the family of generators for this Coxeter group.

        EXAMPLES::

            sage: W = CoxeterGroup(['A', 3], implementation='coxeter3')
            sage: s = W.simple_reflections()
            sage: s[2]*s[1]*s[2]
            [1, 2, 1]
        """
        from sage.sets.family import Family
        return Family(self.index_set(), lambda i: self.element_class(self, [i]))

    gens = simple_reflections

    def from_reduced_word(self, w):
        """
        Return an element of ``self`` from its (reduced) word.

        EXAMPLES::

            sage: W = CoxeterGroup(['A', 3], implementation='coxeter3')
            sage: W.from_reduced_word([1, 3])
            [1, 3]
            sage: W.from_reduced_word([3, 1])
            [1, 3]
        """
        return self.element_class(self, w)

    def rank(self):
        """
        Return the rank of this Coxeter group, that is, the number of generators.

        EXAMPLES::

            sage: W = CoxeterGroup(['A', 3], implementation='coxeter3')
            sage: W.rank()
            3
        """
        return self._coxgroup.rank()

    def is_finite(self):
        """
        Return ``True`` if this is a finite Coxeter group.

        EXAMPLES::

            sage: W = CoxeterGroup(['A', 3], implementation='coxeter3')
            sage: W.is_finite()
            True
        """
        return self._coxgroup.is_finite()

    def length(self, x):
        """
        Return the length of an element ``x`` in this Coxeter group.
        This is just the length of a reduced word for ``x``.

        EXAMPLES::

            sage: W = CoxeterGroup(['A', 3], implementation='coxeter3')
            sage: W.length(W([1,2]))
            2
            sage: W.length(W([1,1]))
            0
        """
        return x.length()

    @cached_method
    def coxeter_matrix(self):
        """
        Return the Coxeter matrix for this Coxeter group.

        The columns and rows are ordered according to the result of
        :meth:`index_set`.

        EXAMPLES::

            sage: W = CoxeterGroup(['A', 3], implementation='coxeter3')
            sage: m = W.coxeter_matrix(); m
            [1 3 2]
            [3 1 3]
            [2 3 1]
            sage: m.index_set() == W.index_set()
            True
        """
        return CoxeterMatrix(self._coxgroup.coxeter_matrix(), self.index_set())

    def root_system(self):
        """
        Return the root system associated with this Coxeter group.

        EXAMPLES::

            sage: W = CoxeterGroup(['A', 3], implementation='coxeter3')
            sage: R = W.root_system(); R
            Root system of type ['A', 3]
            sage: alpha = R.root_space().basis()
            sage: alpha[2] + alpha[3]
            alpha[2] + alpha[3]
        """
        return self.cartan_type().root_system()

    def _an_element_(self):
        """
        Return an element of this Coxeter group.

        EXAMPLES::

            sage: W = CoxeterGroup(['A', 3], implementation='coxeter3')
            sage: W._an_element_()
            []
        """
        return self.element_class(self, [])

    def m(self, i, j):
        r"""
        This is deprecated, use ``self.coxeter_matrix()[i,j]`` instead.

        TESTS::

            sage: W = CoxeterGroup(['A', 3], implementation='coxeter3')
            sage: W.m(1, 1)
            doctest:warning...:
            DeprecationWarning: the .m(i, j) method has been deprecated; use .coxeter_matrix()[i,j] instead.
            See https://github.com/sagemath/sage/issues/30237 for details.
            1
        """
        from sage.misc.superseded import deprecation
        deprecation(30237, "the .m(i, j) method has been deprecated; use .coxeter_matrix()[i,j] instead.")
        return self.coxeter_matrix()[i,j]

    def kazhdan_lusztig_polynomial(self, u, v, constant_term_one=True):
        r"""
        Return the Kazhdan-Lusztig polynomial `P_{u,v}`.

        INPUT:

        - ``u``, ``v`` -- elements of the underlying Coxeter group
        - ``constant_term_one`` -- boolean (default: ``True``); ``True`` uses
          the constant equals one convention, ``False`` uses the Leclerc-Thibon
          convention

        .. SEEALSO::

            - :class:`~sage.combinat.kazhdan_lusztig.KazhdanLusztigPolynomial`
            - :meth:`parabolic_kazhdan_lusztig_polynomial`

        EXAMPLES::

            sage: W = CoxeterGroup(['A', 3], implementation='coxeter3')
            sage: W.kazhdan_lusztig_polynomial([], [1,2, 1])
            1
            sage: W.kazhdan_lusztig_polynomial([1],[3,2])
            0
            sage: W = CoxeterGroup(['A',3],implementation='coxeter3')
            sage: W.kazhdan_lusztig_polynomial([2],[2,1,3,2])
            q + 1

        .. NOTE::

            Coxeter3, as well as Sage's native implementation in
            :class:`~sage.combinat.kazhdan_lusztig.KazhdanLusztigPolynomial`
            use the convention under which Kazhdan-Lusztig
            polynomials give the change of basis from the `(C_w)_{w\in W}`
            basis to the `(T_w)_{w\in W}` of the Hecke algebra of `W` with
            parameters `q` and `q^{-1}`:

                .. MATH:: C_w = \sum_u  P_{u,w} T_u.

            In particular, `P_{u,u}=1`::

                sage: all(W.kazhdan_lusztig_polynomial(u,u) == 1 for u in W)
                True

            This convention differs from Theorem 2.7 in [LT1998]_ by:

            .. MATH::

                {}^{LT} P_{y,w}(q) = q^{\ell(w)-\ell(y)} P_{y,w}(q^{-2})

            To access the Leclerc-Thibon convention use::

                sage: W = CoxeterGroup(['A',3],implementation='coxeter3')
                sage: W.kazhdan_lusztig_polynomial([2],[2,1,3,2],constant_term_one=False)
                q^3 + q

        TESTS:

        We check that Coxeter3 and Sage's implementation give the same results::

            sage: C = CoxeterGroup(['B', 3], implementation='coxeter3')
            sage: W = WeylGroup("B3",prefix='s')
            sage: [s1,s2,s3] = W.simple_reflections()
            sage: R.<q> = LaurentPolynomialRing(QQ)
            sage: KL = KazhdanLusztigPolynomial(W,q)
            sage: all(KL.P(1,w) == C.kazhdan_lusztig_polynomial([],w.reduced_word()) for w in W)  # long (15s)
            True
        """
        u, v = self(u), self(v)
        p = u.value.kazhdan_lusztig_polynomial(v.value)
        if constant_term_one:
            return p
        ZZq = PolynomialRing(ZZ, 'q', sparse=True)
        # This is the same as q**len_diff * p(q**(-2))
        len_diff = v.length()-u.length()
        d = {-2*deg+len_diff: coeff for deg,coeff in enumerate(p) if coeff != 0}
        return ZZq(d)

    def parabolic_kazhdan_lusztig_polynomial(self, u, v, J, constant_term_one=True):
        r"""
        Return the parabolic Kazhdan-Lusztig polynomial `P_{u,v}^{-,J}`.

        INPUT:

        - ``u``, ``v`` -- minimal length coset representatives of `W/W_J` for this Coxeter group `W`
        - ``J`` -- a subset of the index set of ``self`` specifying the parabolic subgroup

        This method implements the parabolic Kazhdan-Lusztig polynomials
        `P^{-,J}_{u,v}` of [Deo1987b]_, which are defined as
        `P^{-,J}_{u,v} = \sum_{z\in W_J} (-1)^{\ell(z)} P_{yz,w}(q)`
        with the conventions in Sage.
        As for :meth:`kazhdan_lusztig_polynomial` the convention
        differs from Theorem 2.7 in [LT1998]_ by:

        .. MATH::

            {}^{LT} P_{y,w}^{-,J}(q) = q^{\ell(w)-\ell(y)} P_{y,w}^{-,J}(q^{-2})

        EXAMPLES::

            sage: W = CoxeterGroup(['A',3], implementation='coxeter3')
            sage: W.parabolic_kazhdan_lusztig_polynomial([],[3,2],[1,3])
            0
            sage: W.parabolic_kazhdan_lusztig_polynomial([2],[2,1,3,2],[1,3])
            q

            sage: C = CoxeterGroup(['A',3,1], implementation='coxeter3')
            sage: C.parabolic_kazhdan_lusztig_polynomial([],[1],[0])
            1
            sage: C.parabolic_kazhdan_lusztig_polynomial([],[1,2,1],[0])
            1
            sage: C.parabolic_kazhdan_lusztig_polynomial([],[0,1,0,1,2,1],[0])
            q
            sage: w=[1, 2, 1, 3, 0, 2, 1, 0, 3, 0, 2]
            sage: v=[1, 2, 1, 3, 0, 1, 2, 1, 0, 3, 0, 2, 1, 0, 3, 0, 2]
            sage: C.parabolic_kazhdan_lusztig_polynomial(w,v,[1,3])
            q^2 + q
            sage: C.parabolic_kazhdan_lusztig_polynomial(w,v,[1,3],constant_term_one=False)
            q^4 + q^2

        TESTS::

            sage: W = CoxeterGroup(['A', 3], implementation='coxeter3')
            sage: type(W.parabolic_kazhdan_lusztig_polynomial([2],[],[1]))
            <class 'sage.rings.polynomial.polynomial_integer_dense_flint.Polynomial_integer_dense_flint'>
        """
        u = self(u)
        v = self(v)
        if any(d in J for d in u.descents()) or any(d in J for d in v.descents()):
            raise ValueError("u and v have to be minimal coset representatives")
        J_set = set(J)
        WOI = self.weak_order_ideal(lambda x: J_set.issuperset(x.descents()))
        if constant_term_one:
            P = PolynomialRing(ZZ, 'q')
            return P.sum((-1)**(z.length()) * self.kazhdan_lusztig_polynomial(u*z,v)
                         for z in WOI if (u*z).bruhat_le(v))
        P = PolynomialRing(ZZ, 'q', sparse=True)
        return P.sum((-1)**(z.length()) * self.kazhdan_lusztig_polynomial(u*z,v, constant_term_one=False).shift(z.length())
                     for z in WOI if (u*z).bruhat_le(v))

    class Element(ElementWrapper):
        wrapped_class = CoxGroupElement

        def __init__(self, parent, x):
            """
            TESTS::

                sage: W = CoxeterGroup(['A', 3], implementation='coxeter3')
                sage: W([2,1,2])
                [1, 2, 1]

            Check that :issue:`32266` is fixed::

                sage: A3 = CoxeterGroup('A3', implementation='coxeter3')
                sage: s1,s2,s3 = A3.simple_reflections()
                sage: s1*s3
                [1, 3]
                sage: s3*s1
                [1, 3]
                sage: s3*s1 == s1*s3
                True
            """
            if not isinstance(x, CoxGroupElement):
                x = CoxGroupElement(parent._coxgroup, x).reduced()
            x = x.normal_form()
            ElementWrapper.__init__(self, parent, x)

        def __iter__(self):
            """
            Return an iterator for the elements in the reduced word.

            EXAMPLES::

                sage: W = CoxeterGroup(['A', 3], implementation='coxeter3')
                sage: w = W([1,2,1])
                sage: list(iter(w))
                [1, 2, 1]
            """
            return iter(self.value)

        def coatoms(self):
            """
            Return the coatoms (or co-covers) of this element in the Bruhat order.

            EXAMPLES::

                sage: W = CoxeterGroup(['B', 3], implementation='coxeter3')
                sage: w = W([1,2,3])
                sage: w.coatoms()
                [[2, 3], [3, 1], [1, 2]]
            """
            W = self.parent()
            return [W(w) for w in self.value.coatoms()]

        def _richcmp_(self, other, op):
            """
            Return lexicographic comparison of ``self`` and ``other``.

            EXAMPLES::

                sage: W = CoxeterGroup(['B', 3], implementation='coxeter3')
                sage: w = W([1,2,3])
                sage: v = W([3,1,2])
                sage: v < w
                False
                sage: w < v
                True

            Some tests for equality::

                sage: W = CoxeterGroup(['A', 3], implementation='coxeter3')
                sage: W([1,2,1]) == W([2,1,2])
                True
                sage: W([1,2,1]) == W([2,1])
                False
            """
            return richcmp(list(self), list(other), op)

        def reduced_word(self):
            """
            Return the reduced word of ``self``.

            EXAMPLES::

                sage: W = CoxeterGroup(['B', 3], implementation='coxeter3')
                sage: w = W([1,2,3])
                sage: w.reduced_word()
                [1, 2, 3]
            """
            return list(self)

        def __invert__(self):
            """
            Return the inverse of this Coxeter group element.

            EXAMPLES::

                sage: W = CoxeterGroup(['A', 3], implementation='coxeter3')
                sage: w = W([1,2,3])
                sage: ~w
                [3, 2, 1]
            """
            return self.__class__(self.parent(), ~self.value)

        inverse = __invert__

        def __getitem__(self, i):
            """
            EXAMPLES::

                sage: W = CoxeterGroup(['A', 3], implementation='coxeter3')
                sage: w0 = W([1,2,1])
                sage: w0[0]
                1
                sage: w0[1]
                2
            """
            # Allow the error message to be raised by the underlying element
            return self.value[i]

        def _mul_(self, y):
            """
            EXAMPLES::

                sage: W = CoxeterGroup(['A', 3], implementation='coxeter3')
                sage: s = W.gens()
                sage: s[1]._mul_(s[1])
                []
                sage: s[1]*s[2]*s[1]
                [1, 2, 1]
                sage: s[2]*s[1]*s[2]
                [1, 2, 1]
            """
            return self.__class__(self.parent(), self.value * y.value)

        def __len__(self):
            """
            EXAMPLES::

                sage: W = CoxeterGroup(['A', 3], implementation='coxeter3')
                sage: w = W([1,2,1])
                sage: w.length()
                3
                sage: len(w)
                3
            """
            return len(self.value)

        length = __len__

        def bruhat_le(self, v):
            r"""
            Return whether ``self`` `\le` ``v`` in Bruhat order.

            EXAMPLES::

                sage: W = CoxeterGroup(['A', 3], implementation='coxeter3')
                sage: W([]).bruhat_le([1,2,1])
                True
            """
            v = self.parent()(v)
            return self.value.bruhat_le(v.value)

        def poincare_polynomial(self):
            """
            Return the Poincaré polynomial associated with this element.

            EXAMPLES::

                sage: W = CoxeterGroup(['A', 2], implementation='coxeter3')
                sage: W.long_element().poincare_polynomial()
                t^3 + 2*t^2 + 2*t + 1
                sage: W = CoxeterGroup(['A', 3], implementation='coxeter3')
                sage: W([2,1,3,2]).poincare_polynomial()
                t^4 + 4*t^3 + 5*t^2 + 3*t + 1
                sage: W([1,2,3,2,1]).poincare_polynomial()
                t^5 + 4*t^4 + 6*t^3 + 5*t^2 + 3*t + 1
                sage: rw = sage.combinat.permutation.from_reduced_word
                sage: p = [w.poincare_polynomial() for w in W]
                sage: [rw(w.reduced_word()) for i,w in enumerate(W) if p[i] != p[i].reverse()]
                [[3, 4, 1, 2], [4, 2, 3, 1]]
            """
            return self.value.poincare_polynomial()

        def has_right_descent(self, i):
            """
            Return whether ``i`` is a right descent of this element.

            EXAMPLES::

                sage: W = CoxeterGroup(['A', 4], implementation='coxeter3')
                sage: W([1,2]).has_right_descent(1)
                False
                sage: W([1,2]).has_right_descent(2)
                True
            """
            return i in self.value.right_descents()

        def has_left_descent(self, i):
            """
            Return ``True`` if ``i`` is a left descent of this element.

            EXAMPLES::

                sage: W = CoxeterGroup(['A', 4], implementation='coxeter3')
                sage: W([1,2]).has_left_descent(1)
                True
                sage: W([1,2]).has_left_descent(2)
                False
            """
            return i in self.value.left_descents()

        def action(self, v):
            """
            Return the action of this Coxeter group element on the root space.

            INPUT:

            - ``v`` -- an element of the root space associated with the Coxeter group for ``self``

            EXAMPLES::

                sage: W = CoxeterGroup(['B', 3], implementation='coxeter3')
                sage: R = W.root_system().root_space()
                sage: v = R.an_element(); v
                2*alpha[1] + 2*alpha[2] + 3*alpha[3]
                sage: w = W([1,2,3])
                sage: w.action(v)
                -alpha[1] + alpha[2] + alpha[3]
            """
            # TODO: Find a better way to do this
            W = self.parent().root_system().root_space().weyl_group()
            w = W.from_reduced_word(list(self))
            return w.action(v)

        def action_on_rational_function(self, f):
            r"""
            Return the natural action of this Coxeter group element on a
            polynomial considered as an element of `S(\mathfrak{h}^*)`.

            .. NOTE::

               Note that the number of variables in the polynomial
               ring must correspond to the rank of this Coxeter
               group. The ordering of the variables is assumed to
               coincide with the result of :meth:`index_set`.

            EXAMPLES::

                sage: W = CoxeterGroup(['A', 3], implementation='coxeter3')
                sage: S = PolynomialRing(QQ, 'x,y,z').fraction_field()
                sage: x,y,z = S.gens()
                sage: W([1]).action_on_rational_function(x+y+z)
                (x^2*y + x*z + 1)/x
                sage: W([2]).action_on_rational_function(x+y+z)
                (x*y^2 + y^2*z + 1)/y
                sage: W([3]).action_on_rational_function(x+y+z)
                (y*z^2 + x*z + 1)/z
            """
            Q = f.parent()
            Q_gens = Q.gens()
            W = self.parent()
            R = W.root_system().root_space()
            alpha = R.basis()
            n = W.rank()

            if Q.ngens() != n:
                raise ValueError("the number of generators for the polynomial "
                                 "ring must be the same as the rank of the "
                                 "root system")

            basis_elements = [alpha[i] for i in W.index_set()]
            basis_to_order = {s: i for i, s in enumerate(W.index_set())}

            results = []
            for poly in [f.numerator(), f.denominator()]:
                result = 0
                exponents = poly.exponents()

                for exponent in exponents:
                    # Construct something in the root lattice from the exponent vector
                    exponent = sum(e*b for e, b in zip(exponent, basis_elements))
                    exponent = self.action(exponent)

                    monomial = 1
                    for s, c in exponent.monomial_coefficients().items():
                        monomial *= Q_gens[basis_to_order[s]]**int(c)

                    result += monomial

                results.append(result)

            numerator, denominator = results
            return numerator / denominator



# distutils: language = c++
# distutils: libraries = coxeter3
# sage_setup: distribution = sagemath-coxeter3
# sage.doctest: optional - coxeter3
"""
Low level part of the interface to Fokko Ducloux's Coxeter 3 library

.. TODO::

    - Write a more efficient method for converting polynomials in
      Coxeter to Sage polynomials.
"""
# ****************************************************************************
#       Copyright (C) 2009-2013 Mike Hansen <mhansen@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.libs.coxeter3.decl cimport *
from cpython.object cimport Py_LT, Py_LE, Py_EQ, Py_NE, Py_GT, Py_GE
from sage.cpython.string cimport str_to_bytes, bytes_to_str

initConstants()

from sage.rings.integer import Integer
from sage.rings.integer_ring import ZZ
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing


cdef class String:
    def __init__(self, s=""):
        """
        Construct a Coxeter string from a Python string.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import String
            sage: s = String("hello"); s
            hello
            sage: del s
        """
        self.x = c_String(str_to_bytes(s))

    def __repr__(self):
        """
        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import String
            sage: s = String('Hi')
            sage: s
            Hi
        """
        return bytes_to_str(self.x.ptr())

    def __hash__(self):
        """
        Return the hash of this String.

        This is the hash of the tuple consisting of the class name and
        the name of this type.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import String
            sage: s = String('hello')
            sage: hash(s) == hash('hello')
            True
        """
        return hash(repr(self))

    def __richcmp__(String self, other, int op):
        """
        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import String
            sage: ta1 = String('A')
            sage: ta2 = String('A')
            sage: tb = String('b')
            sage: ta1 == ta2
            True
            sage: tb != ta1
            True
            sage: all([ta1 < tb, ta1 <= tb, ta1 <= ta1])
            True
            sage: all([tb > ta1, tb >= ta1, tb >= tb])
            True
        """
        if type(other) is not type(self):
            if op in (Py_LT, Py_LE, Py_GT, Py_GE):
                return NotImplemented
            return op == Py_NE

        s = repr(self)
        o = repr(other)

        if op == Py_EQ:
            return s == o
        elif op == Py_NE:
            return s != o
        elif op == Py_LT:
            return s < o
        elif op == Py_LE:
            return s <= o
        elif op == Py_GT:
            return s > o
        elif op == Py_GE:
            return s >= o

    def __len__(self):
        """
        Return the length of this string.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import String
            sage: s = String('Hi')
            sage: len(s)
            2
        """
        return self.x.length()

    def __reduce__(self):
        """
        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import String
            sage: s = String('Hi')
            sage: TestSuite(s).run()
        """
        return (String, (repr(self),) )


cdef class Type:
    def __init__(self, s):
        """
        Construct a Coxeter Type from a Python string.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import Type
            sage: t = Type('A'); t
            A
            sage: del t
        """
        self.x = c_Type(str_to_bytes(s))

    def __repr__(self):
        """
        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import Type
            sage: t = Type('A'); t
            A
        """
        return bytes_to_str(self.x.name().ptr())

    def name(self):
        """
        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import Type
            sage: t = Type('A')
            sage: t.name()
            A
        """
        return String(bytes_to_str(self.x.name().ptr()))

    def __hash__(self):
        """
        Return the hash of this Type.

        This is the hash of the tuple consisting of the class name and
        the name of this type.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import Type
            sage: a = Type('A')
            sage: b = Type('B')
            sage: hash(a) == hash(b)
            False
            sage: d = {a: 1, b: 2}
        """
        return hash(('Type', self.name()))

    def __richcmp__(Type self, other, int op):
        """
        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import Type
            sage: ta1 = Type('A')
            sage: ta2 = Type('A')
            sage: tb = Type('b')
            sage: ta1 == ta2
            True
            sage: tb != ta1
            True
            sage: all([ta1 < tb, ta1 <= tb, ta1 <= ta1])
            True
            sage: all([tb > ta1, tb >= ta1, tb >= tb])
            True
        """
        if type(other) is not type(self):
            if op in (Py_LT, Py_LE, Py_GT, Py_GE):
                return NotImplemented
            return op == Py_NE

        s = repr(self)
        o = repr(other)

        if op == Py_EQ:
            return s == o
        elif op == Py_NE:
            return s != o
        elif op == Py_LT:
            return s < o
        elif op == Py_LE:
            return s <= o
        elif op == Py_GT:
            return s > o
        elif op == Py_GE:
            return s >= o

    def __reduce__(self):
        """
        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import Type
            sage: t = Type('A')
            sage: TestSuite(t).run()
        """
        return (Type, (repr(self), ))


cdef class CoxGroup(SageObject):
    def __cinit__(self, cartan_type):
        """
        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import get_CoxGroup as CoxGroup
            sage: W = CoxGroup(['A', 5]); W
            Coxeter group of type A and rank 5

        Coxeter 3 segfault's on the trivial Coxeter group; so we catch
        this and raise a not implemented error::

            sage: W = CoxGroup(['A', 0]); W
            Traceback (most recent call last):
            ...
            NotImplementedError: Coxeter group of type ['A',0] using Coxeter 3 not yet implemented

        Successfully initializes from a relabeled Cartan type::

            sage: ctype = CartanType(['B', 3]).relabel({1: 3, 2: 2, 3: 1})
            sage: W = CoxGroup(ctype)
            sage: CoxeterMatrix(W.coxeter_matrix(), ctype.index_set()) == CoxeterMatrix(ctype)
            True
        """
        from sage.combinat.root_system.cartan_type import CartanType
        from sage.combinat.root_system.coxeter_matrix import CoxeterMatrix
        self.cartan_type = CartanType(cartan_type)
        ordering = self._ordering_from_cartan_type(self.cartan_type)

        type, rank = self.cartan_type.type(), self.cartan_type.rank()
        if self.cartan_type.is_affine():
            # Only untwisted affine groups are supported
            try:
                if not self.cartan_type.is_untwisted_affine():
                    raise NotImplementedError('twisted affine groups are not supported in coxeter3')
            except AttributeError:
                pass
            type = type.lower()

        type = 'B' if type == 'C' else type

        if rank == 0:
            raise NotImplementedError("Coxeter group of type ['A',0] using Coxeter 3 not yet implemented")
        cdef Type t = Type(type)
        cdef c_CoxGroup* c_W = coxeterGroup(t.x, rank)
        self.x = c_W
        self.out_ordering = {i+1: o for i,o in enumerate(ordering)}
        self.in_ordering = {self.out_ordering[a]: a for a in self.out_ordering}

        # If the Cartan type supplied is relabeled, compose these orderings
        # with the relabelling on the appropriate sides:
        if hasattr(self.cartan_type, '_relabelling'):
            r = self.cartan_type._relabelling
            r_inv = {v: k for (k, v) in r.items()}
            # Pre-compose in_ordering with r
            self.in_ordering = {i: self.in_ordering[r[i]] for i in self.in_ordering}
            # Post-compose out_ordering with r inverse
            self.out_ordering = {i: r_inv[self.out_ordering[i]] for i in self.out_ordering}

        # Check that the Coxeter matrices match up.
        cox_mat = CoxeterMatrix(self.coxeter_matrix(), self.cartan_type.index_set())
        if cox_mat != CoxeterMatrix(self.cartan_type):
            print("Warning, differing Coxeter matrices")

    @classmethod
    def _ordering_from_cartan_type(cls, cartan_type):
        """
        Return an ordering of the index set associated to the Cartan type.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import get_CoxGroup as CoxGroup
            sage: W = CoxGroup(['A', 5])
            sage: W._ordering_from_cartan_type(CartanType(['A',5]))
            [1, 2, 3, 4, 5]
        """
        from sage.arith.srange import srange
        t = cartan_type.type()
        r = cartan_type.rank()
        is_affine = cartan_type.is_affine()

        if t in ['B', 'C', 'D', 'F', 'H']:
            return srange(r-1 if is_affine else r,
                          -1 if is_affine else 0, -1)
        elif t in ['A', 'I']:
            return srange(0 if is_affine else 1, r+1)
        elif t in ['G']:
            if is_affine:
                raise NotImplementedError
            else:
                return [Integer(1), Integer(2)]
        elif t in ['E']:
            if is_affine:
                return srange(1, r) + [ZZ.zero()]
            else:
                return srange(1, r+1)
        else:
            raise NotImplementedError

    def __hash__(self):
        """
        Return the hash of this CoxGroup.

        This is the hash of the tuple of the class's name, the type,
        and the rank.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import get_CoxGroup as CoxGroup
            sage: A4 = CoxGroup(['A', 4])
            sage: d = {A4: True}
        """
        return hash((self.__class__.__name__, self.type(), self.rank()))

    def __richcmp__(CoxGroup self, other, int op):
        """
        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import get_CoxGroup as CoxGroup
            sage: A4 = CoxGroup(['A', 4])
            sage: A5 = CoxGroup(['A', 5])
            sage: B4 = CoxGroup(['B', 4])
            sage: A4 == A4
            True
            sage: A4 != B4
            True
            sage: A4 < B4
            True
            sage: A5 > A4
            True
            sage: A4 >= A4
            True
            sage: B4 >= A5
            True
        """
        if type(other) is not type(self):
            if op in (Py_LT, Py_LE, Py_GT, Py_GE):
                return NotImplemented
            return op == Py_NE

        s_t = self.type()
        o_t = other.type()
        s_r = self.rank()
        o_r = other.rank()

        if op == Py_EQ:
            return s_t == o_t and s_r == o_r
        elif op == Py_NE:
            return s_t != o_t or s_r != o_r
        elif op == Py_LT:
            return s_t < o_t or (s_t == o_t and s_r < o_r)
        elif op == Py_LE:
            return s_t < o_t or (s_t == o_t and s_r <= o_r)
        elif op == Py_GT:
            return s_t > o_t or (s_t == o_t and s_r > o_r)
        elif op == Py_GE:
            return s_t > o_t or (s_t == o_t and s_r >= o_r)

    def __reduce__(self):
        """
        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import get_CoxGroup as CoxGroup
            sage: W = CoxGroup(['A', 5])
            sage: TestSuite((W)).run()
        """
        return (CoxGroup, (self.cartan_type,))

    def __dealloc__(self):
        """
        Deallocate the memory for this CoxGroup.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import get_CoxGroup as CoxGroup
            sage: W = CoxGroup(['A', 5])
            sage: del W
        """
        del self.x

    def __repr__(self):
        """
        Return a string representation of this Coxeter group.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import get_CoxGroup as CoxGroup
            sage: W = CoxGroup(['A', 5]); W
            Coxeter group of type A and rank 5
        """
        return "Coxeter group of type %s and rank %s"%(self.type(), self.rank())

    def __iter__(self):
        """
        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import get_CoxGroup as CoxGroup
            sage: W = CoxGroup(['A', 2])
            sage: list(iter(W))
            [[], [1], [2], [1, 2], [2, 1], [1, 2, 1]]
        """
        return CoxGroupIterator(self)

    def bruhat_interval(self, w, v):
        """
        Return the list of the elements in the Bruhat interval between `w` and `v`.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import get_CoxGroup as CoxGroup
            sage: W = CoxGroup(['A', 2])
            sage: W.bruhat_interval([], [1,2])
            [[], [1], [2], [1, 2]]
        """
        cdef CoxGroupElement ww = CoxGroupElement(self, w)
        cdef CoxGroupElement vv = CoxGroupElement(self, v)
        cdef c_List_CoxWord l = c_List_CoxWord(0)
        interval(l, self.x[0], ww.word, vv.word)
        bruhat_interval = []
        cdef CoxGroupElement u
        cdef CoxGroupElement gg = CoxGroupElement(self, [])
        cdef size_t j
        for j in range(l.size()):
            u = gg._new()
            u.word = l[j]
            bruhat_interval.append(u)

        return bruhat_interval

    def orderings(self):
        """
        Return two dictionaries specifying the mapping of the labels
        of the Dynkin diagram between Sage and Coxeter3 and the its
        inverse.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import get_CoxGroup as CoxGroup
            sage: W = CoxGroup(['A', 5])
            sage: W.orderings()
            ({1: 1, 2: 2, 3: 3, 4: 4, 5: 5}, {1: 1, 2: 2, 3: 3, 4: 4, 5: 5})
        """
        return self.in_ordering, self.out_ordering

    def type(self):
        """
        Return the type of this Coxeter group.

        Note that the type does not include the rank.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import get_CoxGroup as CoxGroup
            sage: W = CoxGroup(['A', 5])
            sage: W.type()
            A
        """
        return Type(bytes_to_str(self.x.type().name().ptr()))

    def rank(self):
        """
        Return the rank of this Coxeter group.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import get_CoxGroup as CoxGroup
            sage: W = CoxGroup(['A', 5])
            sage: W.rank()
            5
        """
        return self.x.rank()

    def order(self):
        """
        Return the order of this Coxeter group.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import get_CoxGroup as CoxGroup
            sage: W = CoxGroup(['A', 5])
            sage: W.order()
            720
            sage: W = CoxGroup(['A', 3, 1])
            sage: W.order()
            +Infinity
        """
        if self.is_finite():
            return Integer(self.x.order())
        else:
            from sage.rings.infinity import infinity
            return infinity

    def is_finite(self):
        """
        Return whether this Coxeter group is finite.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import get_CoxGroup as CoxGroup
            sage: W = CoxGroup(['A', 5])
            sage: W.is_finite()
            True
            sage: W = CoxGroup(['A', 3, 1])
            sage: W.is_finite()
            False
        """
        return isFiniteType(self.x)

    cpdef full_context(self) noexcept:
        """
        Make all of the elements of a finite Coxeter group available.

        Raises an error if ``self`` is not finite.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import get_CoxGroup as CoxGroup
            sage: W = CoxGroup(['A', 2])
            sage: W.full_context()
            sage: W = CoxGroup(['A', 2,1])
            sage: W.full_context()
            Traceback (most recent call last):
            ...
            TypeError: group needs to be finite
        """
        if not self.is_finite():
            raise TypeError("group needs to be finite")
        cdef c_FiniteCoxGroup* fcoxgroup = <c_FiniteCoxGroup*>(self.x)
        if not fcoxgroup.isFullContext():
            fcoxgroup.fullContext()

    def long_element(self):
        """
        Return the longest word in a finite Coxeter group.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import get_CoxGroup as CoxGroup
            sage: W = CoxGroup(['A', 5])
            sage: W.long_element()
            [1, 2, 1, 3, 2, 1, 4, 3, 2, 1, 5, 4, 3, 2, 1]

            sage: W = CoxGroup(['A', 3, 1])
            sage: W.long_element()
            Traceback (most recent call last):
            ...
            TypeError: group needs to be finite
        """
        self.full_context()
        cdef c_FiniteCoxGroup* fcoxgroup = <c_FiniteCoxGroup*>(self.x)
        cdef CoxGroupElement w0 = CoxGroupElement(self, [])
        w0.word = fcoxgroup.longest_coxword()
        return w0

    def __call__(self, w):
        """
        Return a reduced expression for ``w``.

        INPUT:

        - ``w`` -- a word for an element of ``self``, not necessarily reduced

        OUTPUT: a reduced expression for ``w``

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import get_CoxGroup as CoxGroup
            sage: W = CoxGroup(['A', 5])
            sage: w = [1,1,3,5,4,5,4]
            sage: W.__call__(w)
            [3, 4, 5]
        """
        return CoxGroupElement(self, w).reduced()

    def coxeter_matrix(self):
        """
        Return the Coxeter matrix for this Coxeter group.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import get_CoxGroup as CoxGroup
            sage: W = CoxGroup(['A', 5])
            sage: W.coxeter_matrix()
            [1 3 2 2 2]
            [3 1 3 2 2]
            [2 3 1 3 2]
            [2 2 3 1 3]
            [2 2 2 3 1]
        """
        from sage.matrix.constructor import matrix
        rank = self.rank()
        m = matrix(ZZ, rank, rank)
        for i, ii in enumerate(self.cartan_type.index_set()):
            ii = self.in_ordering[ii]-1
            for j, jj in enumerate(self.cartan_type.index_set()):
                jj = self.in_ordering[jj]-1
                m[i,j] = self.x.M(ii, jj)
        return m

    def coxeter_graph(self):
        """
        Return the Coxeter graph for this Coxeter group.

        OUTPUT: a Sage graph

        .. NOTE::

           This uses the labels native to Coxeter3. This is useful
           when trying to obtain the mapping between the labels of
           Sage and Coxeter3.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import get_CoxGroup as CoxGroup
            sage: W = CoxGroup(['A', 5])
            sage: W.coxeter_graph()
            Graph on 5 vertices
            sage: W.coxeter_graph().edges(sort=True)
            [(1, 2, None), (2, 3, None), (3, 4, None), (4, 5, None)]
        """
        from sage.graphs.graph import Graph
        g = Graph()
        m = self.coxeter_matrix()
        rank = self.rank()
        for i, row in enumerate(m.rows()):
            for j in range(i+1,rank):
                if row[j] == 3:
                    g.add_edge(i+1, j+1)
                elif row[j] > 4:
                    g.add_edge(i+1, j+1, row[j])
        return g


cdef class CoxGroupElement:
    def __init__(self, CoxGroup group, w, normal_form=True):
        """
        TESTS::

            sage: from sage.libs.coxeter3.coxeter import get_CoxGroup as CoxGroup, CoxGroupElement
            sage: W = CoxGroup(['A', 5])
            sage: w = CoxGroupElement(W, [2,1,2,1,1], normal_form=False); w
            [2, 1, 2, 1, 1]
            sage: w = CoxGroupElement(W, [1,1,4,5,4], normal_form=False); w
            [1, 1, 4, 5, 4]
            sage: w = CoxGroupElement(W, [1,1,4,5,4]); w
            [4, 5, 4]
            sage: W = CoxGroup(['A', 4])
            sage: CoxGroupElement(W, [1,2,3,2,3])
            [1, 3, 2]
            sage: W = CoxGroup(['A', 4])
            sage: w = CoxGroupElement(W, [1,2,3,2,3])
            sage: del w
        """
        self.group = (<CoxGroup>group).x
        self._parent_group = group
        self.word.reset()
        for i in w:
            self.word.append(self._parent_group.in_ordering[i])

        if normal_form:
            self.group.normalForm(self.word)

    def _coxnumber(self):
        """
        Return the internal integer used by Coxeter3 to represent this element.

        TESTS::

            sage: from sage.libs.coxeter3.coxeter import get_CoxGroup as CoxGroup, CoxGroupElement
            sage: W = CoxGroup(['A', 4])
            sage: w = CoxGroupElement(W, [1,2,3,2,3])
            sage: w._coxnumber()
            7
        """
        return int(self.group.extendContext(self.word))

    def __reduce__(self):
        """
        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import *
            sage: W = CoxGroup(['A',5])
            sage: w = W([1,2,3])
            sage: TestSuite(w).run()
        """
        return (CoxGroupElement, (self._parent_group, list(self)))

    def __invert__(self):
        """
        Return the inverse of this element.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import *
            sage: W = CoxGroup(['A',5])
            sage: w = W([1,2,3])
            sage: ~w
            [3, 2, 1]
        """
        return CoxGroupElement(self._parent_group, reversed(self))

    inverse = __invert__

    cpdef CoxGroup parent_group(self) noexcept:
        """
        Return the parent Coxeter group for this element.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import *
            sage: W = CoxGroup(['A',5])
            sage: w = W([1,2,3])
            sage: w.parent_group()
            Coxeter group of type A and rank 5
        """
        return self._parent_group

    def __getitem__(self, i):
        """
        Return the `i`-th entry of this element.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import *
            sage: W = CoxGroup(['A',5])
            sage: w = W([1,2,3])
            sage: w[0]
            1
            sage: w[2]
            3
            sage: w[:-2]
            [1]
            sage: w[-2:]
            [2, 3]
            sage: w[3:0:-1]
            [3, 2]
            sage: w[4]
            Traceback (most recent call last):
            ...
            IndexError: The index (4) is out of range.
        """
        if isinstance(i, slice):
            #Get the start, stop, and step from the slice
            return [self[ii] for ii in range(*i.indices(len(self)))]
        if i < 0:
            i += len(self)
        if i >= len(self):
            raise IndexError("The index (%d) is out of range." % i)

        return self._parent_group.out_ordering[self.word[i]]

    def __repr__(self):
        """
        Return a string representation of this CoxGroupElement as a list of generators.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import *
            sage: W = CoxGroup(['A',5])
            sage: w = W([1,2,3]); w
            [1, 2, 3]
        """
        return repr(list(self))

    def __hash__(self):
        """
        Return the hash of this element.

        This is a hash of the tuple of the class name, the parent, and
        a tuple of the reduced word.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import *
            sage: W = CoxGroup(['A', 5])
            sage: w = W([1,2,3])
            sage: v = W([2,3,4])
            sage: hash(w) == hash(v)
            False
        """
        return hash((self.__class__.__name__, self.parent_group(), tuple(self)))

    def __richcmp__(CoxGroupElement self, other, int op):
        """
        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import *
            sage: W = CoxGroup(['A', 5])
            sage: V = CoxGroup(['A', 6])
            sage: w1 = W([1,2,3])
            sage: w2 = W([2,3,4])
            sage: v1 = V([1,2,3])
            sage: w1 == w1
            True
            sage: w1 != w2
            True
            sage: all([w1 < w2, w1 <= w2, w1 <= w1])
            True
            sage: all([w2 > w1, w2 >= w1, w2 >= w2])
            True
            sage: w1 == v1
            False
            sage: w1 != v1
            True
        """
        if type(other) is not type(self):
            if op in (Py_LT, Py_LE, Py_GT, Py_GE):
                return NotImplemented
            return op == Py_NE

        s_p = self.parent_group()
        o_p = other.parent_group()
        s_l = list(self)
        o_l = list(other)

        if op == Py_EQ:
            return s_p == o_p and s_l == o_l
        elif op == Py_NE:
            return s_p != o_p or s_l != o_l
        elif op == Py_LT:
            return s_p < o_p or (s_p == o_p and s_l < o_l)
        elif op == Py_LE:
            return s_p < o_p or (s_p == o_p and s_l <= o_l)
        elif op == Py_GT:
            return s_p > o_p or (s_p == o_p and s_l > o_l)
        elif op == Py_GE:
            return s_p > o_p or (s_p == o_p and s_l >= o_l)

    def __iter__(self):
        """
        Return an iterator for the letters in the reduced word for this element.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import *
            sage: W = CoxGroup(['A',5])
            sage: w = W([1,2,3])
            sage: [a for a in w]
            [1, 2, 3]
        """
        return (self[i] for i in range(len(self)))

    def __len__(self):
        """
        Return the length of this element.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import *
            sage: W = CoxGroup(['A',5])
            sage: w = W([1,2,3])
            sage: len(w)
            3
        """
        return self.word.length()

    def left_descents(self):
        """
        Return the left descent set of this element.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import *
            sage: W = CoxGroup(['A',5])
            sage: w = W([1,2,1])
            sage: w.left_descents()
            [1, 2]
        """
        return LFlags_to_list(self._parent_group, self.group.ldescent(self.word))

    def right_descents(self):
        """
        Return the right descent set of this element.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import *
            sage: W = CoxGroup(['A',5])
            sage: w = W([1,2,1])
            sage: w.right_descents()
            [1, 2]
        """
        return LFlags_to_list(self._parent_group, self.group.rdescent(self.word))

    def bruhat_le(self, w):
        """
        Return whether u = (self) is less than w in Bruhat order.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import *
            sage: W = CoxGroup(['A',5])
            sage: w = W([1,2,3,4,5,4])
            sage: v = W([1,2,4,5,4])
            sage: v.bruhat_le(w)
            True
            sage: w.bruhat_le(w)
            True
            sage: w.bruhat_le(v)
            False
        """
        cdef CoxGroupElement ww = CoxGroupElement(self._parent_group, w)
        return self.group.inOrder(self.word, ww.word)

    def is_two_sided_descent(self, s):
        """
        Return whether ``s`` is a two-sided descent of ``self``.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import *
            sage: W = CoxGroup(['A',2])
            sage: x = W([1,2,1])
            sage: x.is_two_sided_descent(1)
            True
        """
        cdef Generator ss = self._parent_group.in_ordering[s]
        return self.group.isDescent(self.word, s)

    cdef CoxGroupElement _new(self) noexcept:
        """
        Return a new copy of this element.
        """
        cdef CoxGroupElement res = CoxGroupElement(self.parent_group(), [])
        res.word = self.word
        return res

    def coatoms(self):
        """
        Return the coatoms of this element in Bruhat order.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import *
            sage: W = CoxGroup(['A',2])
            sage: W([1,2,1]).coatoms()
            [[2, 1], [1, 2]]
            sage: W([]).coatoms()
            []
        """
        cdef c_List_CoxWord list = c_List_CoxWord(0)
        self.group.coatoms(list, self.word)

        coatoms = []

        cdef Length i = 0
        cdef CoxGroupElement res
        for i in range(list.size()):
            res = self._new()
            res.word = list[i]
            coatoms.append(res)
        return coatoms

    def normal_form(self):
        """
        Return ``self`` in normal form.

        This is the lexicographically minimal reduced word for
        ``self``.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import get_CoxGroup as CoxGroup, CoxGroupElement
            sage: W = CoxGroup(['A', 5])
            sage: w = CoxGroupElement(W, [2,1,2], normal_form=False); w
            [2, 1, 2]
            sage: w.normal_form()
            [1, 2, 1]
        """
        cdef CoxGroupElement res = self._new()
        self.group.normalForm(res.word)
        return res

    def reduced(self):
        """
        Return a reduced word for this element.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import get_CoxGroup as CoxGroup, CoxGroupElement
            sage: W = CoxGroup(['A', 5])
            sage: w = CoxGroupElement(W, [2,1,2,1,1], normal_form=False); w
            [2, 1, 2, 1, 1]
            sage: w.reduced()
            [1, 2, 1]
        """
        cdef CoxGroupElement res = self._new()
        self.group.reduced(res.word, self.word)
        return res

    def __mul__(CoxGroupElement self, CoxGroupElement y):
        """
        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import get_CoxGroup as CoxGroup
            sage: W = CoxGroup(['A', 5])
            sage: W([1]) * W([1])
            []
            sage: W([1,2]) * W([1])
            [1, 2, 1]
        """
        cdef CoxGroupElement res = self._new()
        self.group.prod(res.word, y.word)
        return res

    def poincare_polynomial(self):
        """
        Return the Poincaré polynomial associated with the Bruhat
        interval between the identity element and this one.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import get_CoxGroup as CoxGroup
            sage: W = CoxGroup(['A', 5])
            sage: W([]).poincare_polynomial()
            1
            sage: W([1,2,1]).poincare_polynomial()
            t^3 + 2*t^2 + 2*t + 1
        """
        cdef CoxGroup W = self.parent_group()
        cdef c_List_CoxWord result = c_List_CoxWord(0)
        cdef CoxGroupElement id = CoxGroupElement(W, [])
        cdef CoxGroupElement ww = CoxGroupElement(W, self)
        interval(result, W.x[0], id.word, ww.word)

        cdef list coefficients = [0]*(len(ww)+1)
        cdef size_t j
        for j in range(result.size()):
            coefficients[result[j].length()] += 1
        return ZZ['t'](coefficients)

    def kazhdan_lusztig_polynomial(self, v):
        """
        Return the Kazhdan-Lusztig polynomial `P_{u,v}` where `u` is ``self``.

        Currently this is a bit inefficient as it constructs the
        polynomial from its string representation.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import get_CoxGroup as CoxGroup
            sage: W = CoxGroup(['A', 2])
            sage: W([]).kazhdan_lusztig_polynomial([1,2,1])
            1
            sage: W([1,2,1]).kazhdan_lusztig_polynomial([])
            0
        """
        cdef CoxGroupElement vv
        if not isinstance(v, CoxGroupElement):
            vv = CoxGroupElement(self._parent_group, v)
        else:
            vv = v

        ZZq = PolynomialRing(ZZ, 'q')
        if not self.group.inOrder(self.word, vv.word):
            return ZZq.zero()

        cdef CoxNbr x = self.group.extendContext(self.word)
        cdef CoxNbr y = self.group.extendContext(vv.word)
        cdef c_KLPol kl_poly = self.group.klPol(x, y)
        if kl_poly.isZero():
            return ZZq.zero()
        cdef size_t i
        l = [kl_poly[i] for i in range(kl_poly.deg()+1)]
        return ZZq(l)

    def mu_coefficient(self, v):
        r"""
        Return the mu coefficient `\mu(u,v)` where `u` is this element.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import *
            sage: W = CoxGroup(['A',5])
            sage: w = W([1,2,3,4,5,4])
            sage: v = W([1,2,4,5,4])
            sage: w.mu_coefficient(v)
            0
            sage: w.mu_coefficient(w)
            0
            sage: v.mu_coefficient(w)
            1
        """
        cdef CoxGroupElement vv = CoxGroupElement(self._parent_group, v)
        cdef CoxNbr x = self.group.extendContext(self.word)
        cdef CoxNbr y = self.group.extendContext(vv.word)
        return ZZ(self.group.mu(x,y))


cdef LFlags_to_list(CoxGroup parent, LFlags f) noexcept:
    """
    Return the right descent set of this element.

    EXAMPLES::

        sage: from sage.libs.coxeter3.coxeter import *
        sage: W = CoxGroup(['A',5])
        sage: w = W([1,2,1])
        sage: w.right_descents()
        [1, 2]
    """
    cdef Generator s
    cdef LFlags f1 = f
    l = []
    while f1:
        s = firstBit(f1)
        l.append(parent.out_ordering[s+1])
        f1 = f1 & (f1-1)
    return l


class CoxGroupIterator():
    def __init__(self, group):
        """
        A class used to iterate over all of the elements of a Coxeter group.

        .. NOTE::

           This will construct all of the elements of the group within
           Coxeter3.  For some groups, this may be too large to fit
           into memory.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import get_CoxGroup as CoxGroup, CoxGroupIterator
            sage: W = CoxGroup(['A', 2])
            sage: it = CoxGroupIterator(W)
            sage: [next(it) for i in range(W.order())]
            [[], [1], [2], [1, 2], [2, 1], [1, 2, 1]]
        """
        self.group = group
        self.order = group.order()
        self.n = 0
        self.group.full_context()

    def __iter__(self):
        """
        Return self, as per the iterator protocol.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import get_CoxGroup as CoxGroup, CoxGroupIterator
            sage: W = CoxGroup(['A', 2])
            sage: it = iter(W)
            sage: it is iter(it)
            True
        """
        return self

    def __next__(self):
        """
        Return the next element in the associated Coxeter group.

        EXAMPLES::

            sage: from sage.libs.coxeter3.coxeter import get_CoxGroup as CoxGroup, CoxGroupIterator
            sage: W = CoxGroup(['A', 2])
            sage: it = CoxGroupIterator(W)
            sage: next(it)
            []
        """
        if self.n >= self.order:
            raise StopIteration
        cdef CoxGroupElement w = self.group([])

        (<CoxGroup>self.group).x.prod_nbr(w.word, self.n)
        self.n += 1
        return w

    next = __next__


CoxGroup_cache = {}


def get_CoxGroup(cartan_type):
    """
    TESTS::

        sage: from sage.libs.coxeter3.coxeter import get_CoxGroup as CoxGroup, CoxGroupIterator
        sage: W = CoxGroup(['A', 2])
    """
    from sage.combinat.root_system.cartan_type import CartanType
    cartan_type = CartanType(cartan_type)
    if cartan_type not in CoxGroup_cache:
        CoxGroup_cache[cartan_type] = CoxGroup(cartan_type)
    return CoxGroup_cache[cartan_type]



#!/usr/bin/env python

from distutils import log
from setuptools import setup

# Work around a Cython problem in Python 3.8.x on macOS
# https://github.com/cython/cython/issues/3262
import os
if os.uname().sysname == 'Darwin':
    import multiprocessing
    multiprocessing.set_start_method('fork', force=True)

# If build isolation is not in use and setuptools_scm is installed,
# then its file_finders entry point is invoked, which we don't need.
# Workaround from ​https://github.com/pypa/setuptools_scm/issues/190#issuecomment-351181286
try:
    import setuptools_scm.integration
    setuptools_scm.integration.find_files = lambda _: []
except ImportError:
    pass

# PEP 517 builds do not have . in sys.path
import sys
sys.path.insert(0, os.path.dirname(__file__))

if len(sys.argv) > 1 and (sys.argv[1] == "sdist" or sys.argv[1] == "egg_info"):
    sdist = True
else:
    sdist = False

if sdist:
    cmdclass = {}
else:
    from sage_setup.excepthook import excepthook
    sys.excepthook = excepthook

    from sage_setup.setenv import setenv
    setenv()

    import sage.env
    sage.env.default_required_modules = sage.env.default_optional_modules = ()

    from sage_setup.command.sage_build_cython import sage_build_cython
    from sage_setup.command.sage_build_ext import sage_build_ext
    from sage_setup.command.sage_build_py import sage_build_py
    sage_build_cython.built_distributions = ['sagemath-coxeter3']

    cmdclass = dict(build_cython=sage_build_cython,
                    build_ext=sage_build_ext,
                    build_py=sage_build_py)

from sage_setup.find import find_python_sources
python_packages, python_modules, cython_modules = find_python_sources(
    '.', ['sage'], distributions=['sagemath-coxeter3'])

log.warn('python_packages = {0}'.format(python_packages))
log.warn('python_modules = {0}'.format(python_modules))
log.warn('cython_modules = {0}'.format(cython_modules))

setup(
    cmdclass = cmdclass,
    packages = python_packages,
    py_modules  = python_modules,
    ext_modules = cython_modules,
)
